:root {
    --card-menu-top: 5px;
    --card-menu-right: 5px;
}
.card-menu {
    position: absolute;
    top: var(--card-menu-top);
        right: var(--card-menu-right);
}
.react-grid-layout {
    min-height: 50px;
}